import argparse
import warnings
from pandas import DataFrame, concat
import os
import shutil

from tensorflow.keras.models import model_from_json

try:
    from functions import *
    from configs import hyper_conf, accept_threshold_for_loss_diff, parameter_tuning_trials
    from data_access import *
except Exception as e:
    from .functions import *
    from .configs import hyper_conf, accept_threshold_for_loss_diff, parameter_tuning_trials
    from .data_access import *


def model_from_to_json(path=None, weights_path=None, model=None, is_writing=False, lr=None):
    if is_writing:
        model_json = model.to_json()
        with open(path, "w") as json_file:
            json_file.write(model_json)
        model.save_weights(weights_path)
    else:
        json_file = open(path, 'r')
        loaded_model_json = json_file.read()
        json_file.close()
        model = model_from_json(loaded_model_json)
        try:
            model.load_weights(weights_path)
        except Exception as e:
            model.load_weights(weights_path)
        return model


def updating_hyper_parameters_related_to_data():
    return None


def get_params(params, comb):
    count = 0
    for p in params:
        _p = type(params[p])(comb[count])
        params[p] = _p
        count += 1
    return params


class NextPurchaseModelPrediction:
    def __init__(self,
                 temp_data_path,
                 max_date,
                 future_date,
                 time_period,
                 directory,
                 indicators
                 ):
        self.temp_data = pd.read_csv(temp_data_path)
        self.max_date = convert_str_to_day(max_date)
        self.future_date = convert_str_to_day(future_date)
        self.directory = directory
        self.customer_indicator, self.amount_indicator, self.time_indicator = indicators.split("*")
        self.customers = list(self.temp_data[self.customer_indicator].unique())
        self.params = hyper_conf('next_purchase') \
            if check_for_existing_parameters(self.directory,'next_purchase') is None else \
            check_for_existing_parameters(self.directory, 'next_purchase')
        self.time_period = time_period
        self.get_actual_value = lambda _min, _max, _value: ((_max - _min) * _value) + _min
        self.prev_model_date = check_model_exists(self.directory, "trained_next_purchase_model", self.time_period)
        self.model = model_from_to_json(path=model_path(self.directory,
                                                        "trained_next_purchase_model",
                                                        self.prev_model_date if self.prev_model_date is not None else get_current_day(),
                                                        self.time_period),
                                        weights_path=weights_path(self.directory,
                                                                  "trained_next_purchase_model",
                                                                  self.prev_model_date if self.prev_model_date is not None else get_current_day(),
                                                                  self.time_period), lr=self.params['lr'])

    def prediction_date_add(self, data, pred_data, pred):
        max_date = max(data[self.time_indicator]) if len(pred_data) == 0 else max(pred_data[self.time_indicator])
        return max_date + datetime.timedelta(days=int(round(pred)))

    def prediction_condition_1(self, data, _pred_actual):
        accept = False
        try:
            if self.max_date < self.prediction_date_add(data, pd.DataFrame(), (_pred_actual * 3) + 0.05):
                accept = True
        except: accept = False
        return accept

    def prediction_condition_2(self, _pred_actual, _pred):
        try: return True if int(round(_pred_actual)) != 0 and _pred > 0 else False
        except: return False

    def predicted_date_in_range_decision(self, end, _pred_data, _predicted_date, customer, _pred_actual, _pred):
        if _predicted_date < end:
            _pred_data = concat([_pred_data, DataFrame([{self.time_indicator: _predicted_date,
                                                         self.customer_indicator: customer,
                                                         'time_diff': _pred_actual,
                                                         'time_diff_norm': _pred}])])
        return _pred_data

    def calculate_prediction(self, data, _pred_data, user_min, user_max):
        x = data_for_customer_prediction(data, _pred_data, self.params)
        try: _pred = self.model.predict(x)[0][-1]
        except: _pred = 0
        _pred_actual = self.get_actual_value(_min=user_min, _max=user_max, _value=_pred)
        return _pred_actual, _pred

    def prediction_per_customer(self,  customer):
        warnings.simplefilter("ignore")
        data = self.temp_data[self.temp_data[self.customer_indicator] == customer]
        user_min, user_max = [list(data[metric])[0] for metric in ['user_min', 'user_max']]
        _pred_data, prediction_data[customer] = DataFrame(), DataFrame()
        _predicted_date = self.future_date - datetime.timedelta(days=1)
        _pred_actual, _pred = self.calculate_prediction(data, _pred_data, user_min, user_max)
        if self.prediction_condition_2(_pred_actual, _pred) and self.prediction_condition_1(data, _pred_actual):
            counter = 0
            while self.max_date < _predicted_date < self.future_date:
                try:
                    _pred_actual, _pred = self.calculate_prediction(data, _pred_data, user_min, user_max)
                    if self.prediction_condition_2(_pred_actual, _pred) and counter < 190:
                        counter += 1
                        _predicted_date = self.prediction_date_add(data, _pred_data, _pred_actual)
                        _pred_data = self.predicted_date_in_range_decision(
                            self.future_date, _pred_data, _predicted_date, customer, _pred_actual, _pred)
                    else: _predicted_date = self.future_date + datetime.timedelta(days=1)
                except Exception as e: _predicted_date = self.future_date + datetime.timedelta(days=1)

            if len(_pred_data) != 0:
                try: _pred_data = _pred_data[(_pred_data[self.time_indicator] >= self.max_date) &
                                             (_pred_data[self.time_indicator] < self.future_date) &
                                             (_pred_data[self.time_indicator] == _pred_data[self.time_indicator])]
                except Exception as e: print(e)
            prediction_data[customer] = _pred_data
        del data, user_max, user_min

    def prediction_execute(self):
        print("*"*5, "PREDICTION", 5*"*")
        print("number of users :", len(self.customers))
        self.temp_data[self.time_indicator] = self.temp_data[self.time_indicator] .apply(lambda x: convert_str_to_day(x))
        self.temp_data = self.temp_data.sort_values(by=[self.customer_indicator, self.time_indicator], ascending=True)
        global prediction_data
        prediction_data = {}
        execute_parallel_run(self.customers, self.prediction_per_customer, arguments=None, parallel=4)
        _result = []
        for c in self.customers:
            try: _result.append(prediction_data[c])
            except Exception as e: print(c)
        pd.concat(_result).to_csv(
            join(self.directory, "temp_next_purchase_results", str(self.customers[0] + "_data.csv")), index=False)


if __name__ == '__main__':
    """
                -P    temp_data_path
                -MD   max_date
                -FD   future_date
                -TP   time_period
                -D    directory
                -IND  customer_indicator, amount_indicator, time_indicator
    """
    parser = argparse.ArgumentParser()
    parser.add_argument("-P", "--temp_data_path", type=str,
                        help="""
                                train, prediction
                        """,
                        )
    parser.add_argument("-C", "--customers", type=str,
                        help="""
                                user_1*user_2*user_4*user_5* ...
                        """,
                        )
    parser.add_argument("-MD", "--max_date", type=str,
                        help="""
                                2021-05-05 (string)
                        """,
                        )
    parser.add_argument("-FD", "--future_date", type=str,
                        help="""
                                2021-11-05 (string)
                        """,
                        )
    parser.add_argument("-D", "--directory", type=str,
                        help="""
                                    /../../..
                            """,
                        )
    parser.add_argument("-TP", "--time_period", type=str,
                        help="""
                                    week, day, 6*months, quarter, ..
                            """,
                        )

    parser.add_argument("-IND", "--indicators", type=str,
                        help="""
                                    customer_indicator*amount_indicator*time_indicator
                            """,
                        )

    arguments = parser.parse_args()
    args = {'temp_data_path': arguments.temp_data_path,
            'max_date': arguments.max_date,
            'future_date': arguments.future_date,
            'directory': arguments.directory,
            'time_period': arguments.time_period,
            'indicators': arguments.indicators}
    prediction = NextPurchaseModelPrediction(**args)
    prediction.prediction_execute()
    del prediction






from os.path import join
import threading
import pandas as pd

try:
    from main import main
    from data_access import GetData
    from utils import get_folder_path, write_yaml, read_yaml
    from configs import conf
    from dashboard import create_dahboard
    from functions import check_for_previous_predicted_clv_results
except Exception as e:
    from .main import main
    from .data_access import GetData
    from .utils import get_folder_path, write_yaml, read_yaml
    from .configs import conf
    from .dashboard import create_dashboard
    from .functions import check_for_previous_predicted_clv_results, check_model_exists
    from .functions import check_for_previous_predicted_clv_results


class CLV:
    """
    job                 :  Train, Prediction. train process is related to creating a model
    order_count         :  it allows us to create feature set of purchase amount model. if it is not assigned (it is not a
                           required argument in order to initialize the clv prediction), platform handles for decide
                           optimum order_count.
     customer_indicator :  This parameter indicates which column represents a unique customer identifier on given data.
     amount_indicator   :  This parameter indicates which column represents purchase value (integer, float ..) on the given data.
     time_indicator     :  This parameter indicates which column represents order checkout date with date format (timestamp)
                           (YYYY/MM/DD hh:mm:ss, YYYY-MM-DD hh:mm:ss, YYYY-MM-DD) on given data.
     date               :  This allows us to query the data with a date filter. This removes data that occurs after the given date.
     data_source        :  The location where the data is stored or the query (check data source for details).
     data_query_path    :  Type of data source to import data to the platform (optional Ms SQL, PostgreSQL,
                           AWS RedShift, Google BigQuery, csv, json, pickle).
     connector          :  if there is a connection parameters as user, pasword, host port, this allows us to assign it
                           as dictionary format (e.g {"user":  , "pw":  *}).
     export_path        :  Export path where the outputs are stored. created models (.json format),
                           tuned parameters (test_parameters.yaml).
                           result data with predicted values per user per predicted order
                           (.csv format) are willing to store at given path.
     time_period        :  A period of time which is willing to predict.
                           Supported time periods 6*month, quarter, month, week, 2*week, '2*month'(Required).
    """
    def __init__(self,
                 customer_indicator=None,
                 amount_indicator=None,
                 time_indicator=None,
                 job=None,
                 date=None,
                 order_count=None,
                 data_source=None,
                 data_query_path=None,
                 time_schedule=None,
                 time_period=None,
                 export_path=None,
                 connector=None):

        self.job = job.lower() if job is not None else job
        self.time_period = time_period.lower() if data_source is not None else data_source
        self.amount_indicator = amount_indicator
        self.order_count = order_count
        self.customer_indicator = customer_indicator
        self.data_source = data_source.lower() if data_source is not None else data_source
        self.data_query_path = data_query_path
        self.data_query_path_raw = data_query_path
        self.time_indicator = time_indicator
        self.time_schedule = time_schedule.lower() if time_schedule is not None else time_schedule
        self.export_path = export_path
        self.connector = connector
        self.result_columns = [customer_indicator, 'order_seq_num', time_indicator, amount_indicator, 'data_type']
        self.raw_data = pd.DataFrame()
        self.results = pd.DataFrame()
        self.sorting_columns = [self.customer_indicator, self.time_indicator]
        self.arguments = {"job": job,
                          "time_period": time_period,
                          "order_count": order_count,
                          "date": date,
                          "customer_indicator": customer_indicator,
                          "amount_indicator": amount_indicator,
                          "data_source": data_source,
                          "data_query_path": data_query_path,
                          "time_indicator": time_indicator,
                          "export_path": export_path}
        self.arg_terminal = {"job": "J",
                             "time_period": "TP",
                             "order_count": "OC",
                             "date": "D",
                             "customer_indicator": "CI",
                             "amount_indicator": "AI",
                             "data_source": "DS",
                             "data_query_path": "DQP",
                             "time_indicator": "TI",
                             "export_path": "EP"}
        self.schedule_arg = "TS"
        self.args_str = ""
        self.mandetory_arguments = ["data_source", "data_query_path", "customer_indicator",
                                    "amount_indicator", "time_indicator", "export_path"]
        self.clv_predicted = None
        self.path = get_folder_path()
        self.model_count = 3

    def query_string_change(self):
        """
        When query with SQL syntax, it is more accurate to sent query string as argument with coverting spaces with "+"
        character.
        """
        if self.data_source in ['mysql', 'postgresql', 'awsredshift', 'googlebigquery']:
            self.data_query_path = self.data_query_path.replace("\r", " ").replace("\n", " ").replace(" ", "+")

    def get_connector(self):
        """
        Connection checks for given data_source, data_path, connection parameters.
        This will be stored at configs.yaml at docs folder.
        If it can read the data well, returns True
        If data is directly imported from any file with .csv, .json and .yaml format, returns True
        :return: True or False
        """
        config = conf('config')
        try:
            if self.data_source not in ["csv", "json"]:
                for i in config['db_connection']:
                    if i != 'data_source':
                        config['db_connection'][i] = self.connector[i]
                    else:
                        config['db_connection']['data_source'] = self.data_source
            write_yaml(join(self.path, "docs"), "configs.yaml", config, ignoring_aliases=False)
            source = GetData(data_source=self.data_source,
                             data_query_path=self.data_query_path,
                             time_indicator=self.time_indicator,
                             feature=self.amount_indicator)
            source.get_connection()
            return True
        except Exception as e:
            print(e)
            if self.data_source not in ["csv", "json"]:
                for i in config['db_connection']:
                    if i is not 'data_source':
                        config['db_connection'][i] = None
                    else:
                        config['db_connection']['data_source'] = self.data_source
            write_yaml(join(self.path, "docs"), "configs.yaml", config, ignoring_aliases=False)
            return False

    def check_for_mandetory_arguments(self):
        """
        checking for mandetory arguments before initializing for model train or prediction process.
        :return: True, False
        """
        for arg in self.arg_terminal:
            if arg in self.mandetory_arguments:
                return False if self.arguments[arg] is None else True

    def check_for_time_period(self):
        """
        When clv_prediction is triggered, valid time period must be checked.
        Here are the valid time periods that platform is supporting;
            - "day", "year", "month", "week", "2*week", '6*month',  "quarter"
        """
        if self.time_period is None:
            return True
        else:
            if self.time_period in ["month", "week", "2*week", '6*month',  "quarter", '2*month']:
                return True
            else: return False

    def checking_for_prediction_process(self):
        """
        When the platform is triggered for prediction, first, trained model for related time_period and time
        must be check. If the current trained models are not valid for this prediction process, it won`t be started.
            Checking models;
                - trained_next_purchase_model
                - trained_purchase_amount_model
                - trained_newcomers_model

            if all models are stored into the export_path, prediction process are able to be calculated.
        """
        accept = False
        if self.job == 'prediction':
            models = [check_model_exists(self.export_path, m, self.time_period)
                      for m in ["trained_next_purchase_model", "trained_purchase_amount_model",
                                "trained_newcomers_model"]]
            if len([1 for model in models if len(model) != 0]) == self.model_count:
                accept = True
        else:
            accept = True
        return accept

    def clv_prediction(self):
        """
        This the process where all has begun.
        Train process is triggered with running clv_prediction.
        Prediction process is triggered with running clv_prediction.
        There are conditions that must be acceoted in order to initialized the process;
            - connection .... Done (get_connector)
            - time period check ... Done (check_for_time_period)
            - required arguments .. Done (check_for_mandetory_arguments)
            - Check for trained model exits if it is the prediction process .. Done(checking_for_prediction_process)
            Sending arguments to main function at main.py which handles the model,
            prediction Deep Learning and other staff
        """
        self.query_string_change()
        if self.get_connector():
            if self.check_for_time_period():
                if self.check_for_mandetory_arguments():
                    if self.checking_for_prediction_process():
                        self.clv_predicted = main(**self.arguments)
                    else:
                        print("Execution : *** Prediction")
                        print("None of trained model has been detected!. Please run for Train process!")
                        print("if these are trained model with .json format please check any of these models of ",
                              "the time period ('month', 'week', ..) matchs with given time preiod.",
                              " Given time period is :", self.time_period)
                        print("argument; job: 'train'")
                else:
                    print("check for the required paramters to initialize CLV Prediction:")
                    print(" - ".join(self.mandetory_arguments))
            else:
                print("optional time periods are :")
                print("month", "week", "2*week", '2*month', "quarter", '6*month')
        else:
            print("pls check for data source connection / path / query.")

    def check_for_raw_data(self):
        """
        if model has been initialized before triggering for raw data,
        raw data can be collected from clv_predicted.next_purchase.data
        However, order_seq_num must be created.
        """
        if self.clv_predicted is not None:
            self.raw_data = self.clv_predicted['next_purchase']['data']
            self.raw_data['data_type'] = 'actual'
            self.raw_data['order_seq_num'] = self.raw_data.sort_values(by=self.sorting_columns).groupby(
                [self.customer_indicator]).cumcount() + 1

    def check_for_result_data(self):
        """
        if model has been initialized before triggering for result data,
        result data can be collected from clv_predicted.purchase_amount.results
        """
        if self.clv_predicted is not None:
            self.results = pd.concat([self.clv_predicted['purchase_amount']['results'],
                                      self.clv_predicted['newcomers']['results']])

    def check_for_result_data_from_previous_progresses(self):
        """
        If model has not been triggered but there is predicted result_Data.csv file in 'export_path'
        it can be collected from the given path by merging with previous available result_data.csv files.
        """
        if self.clv_predicted is None or len(self.results) == 0:
            self.results = check_for_previous_predicted_clv_results(self.results,
                                                                    self.export_path,
                                                                    self.time_period,
                                                                    self.time_indicator,
                                                                    self.customer_indicator,
                                                                    )

    def get_result_data(self):
        """
        When model prediction has been done, this allows us to get the predicted data with merge to the raw data.
        If model has not been initialized (clv_prediction has not been run yet),
        it directly collects the data from last predicted result_data.csv
        with given arguments (time_period, time_indicator, export_path, customer_indicator)
        :return: data frame
        """
        self.check_for_raw_data()
        self.check_for_result_data()
        self.check_for_result_data_from_previous_progresses()
        return self.results if self.clv_predicted is None else pd.concat([self.raw_data, self.results])

    def show_dashboard(self):
        """
        if you are running dashboard, make sure you have assigned export_path.
        """
        process = threading.Thread(target=create_dashboard,
                                   kwargs={"customer_indicator": self.customer_indicator,
                                           "amount_indicator": self.amount_indicator,
                                           "directory": self.export_path,
                                           "time_indicator": self.time_indicator,
                                           "time_period": self.time_period,
                                           "data_query_path": self.data_query_path_raw,
                                           "data_source": self.data_source})
        process.daemon = True
        process.start()

    def create_api_for_runnig_realtime_customer_value_predicion(self):
        """
        Real time api for prediction values for individual customer or customer list
        :return:
        """




import argparse

try:
    from next_purchase_model import TrainLSTM
    from purchase_amount_model import TrainConv1Dimension
    from newcomers import TrainLSTMNewComers
except Exception as e:
    from .next_purchase_model import TrainLSTM
    from .purchase_amount_model import TrainConv1Dimension
    from .newcomers import TrainLSTMNewComers


def main(job='train',
         order_count=None,
         customer_indicator=None,
         amount_indicator=None,
         data_source=None,
         data_query_path=None,
         time_period=None,
         time_indicator=None,
         export_path=None,
         date=None):
    """
    This is how we run the model prediction and train processes.
    Depending on the job it is triggered for train and prediction process individually.
    job, customer_indicator, amount_indicator, data_source, data_query_path, time_period, export_path are required.
    if **order_count** is null, it is detecting for optimum order count.
    the date is crucial for data query. (< date)

    :param job: train or prediction
    :param order_count: number of order for purchase amount model
    :param customer_indicator: customer column at data
    :param amount_indicator: amount column at data
    :param data_source: postgres, .csv, .json, awsredshift, bigquery.
    :param data_query_path: query or file_name (with whole path).
    :param time_period: a period of time which is willing to predict.
    :param time_indicator: time columns at data set
    :param export_path: where data, model,tuned_parameters, schedule arguments are stored
    :param date: given date of previous data (< date)
    :return: class TrainLSTM (Next Purchase Model) & class TrainConv1Dimension (Purchase Amount Model)
    """
    print("received :", {"job": job,
                         'order_count': order_count,
                         'customer_indicator': customer_indicator,
                         'amount_indicator': amount_indicator,
                         'data_source': data_source,
                         'data_query_path': data_query_path,
                         'time_period': time_period,
                         'time_indicator': time_indicator,
                         'export_path': export_path}
          )

    # engaged customers of CLV models train and prediction process
    next_purchase = TrainLSTM(
        date=date,
        time_indicator=time_indicator,
        order_count=order_count,
        data_source=data_source,
        data_query_path=data_query_path,
        time_period=time_period,
        directory=export_path,
        customer_indicator=customer_indicator,
        amount_indicator=amount_indicator)
    if job == 'train':
        next_purchase.train_execute()
    if job == 'prediction':
        next_purchase.prediction_execute()
    if job == 'train_prediction':
        next_purchase.train_execute()
        next_purchase.prediction_execute()

    predicted_orders = next_purchase.results
    raw_data = next_purchase.data
    del next_purchase

    purchase_amount = TrainConv1Dimension(
        date=date,
        time_indicator=time_indicator,
        order_count=order_count,
        data_source=data_source,
        data_query_path=data_query_path,
        time_period=time_period,
        directory=export_path,
        customer_indicator=customer_indicator,
        predicted_orders=predicted_orders,
        amount_indicator=amount_indicator)

    if job == 'train':
        purchase_amount.train_execute()
    if job == 'prediction':
        purchase_amount.prediction_execute()
    if job == 'train_prediction':
        purchase_amount.train_execute()
        purchase_amount.prediction_execute()

    engaged_customers_results = purchase_amount.results
    del purchase_amount

    # newcomers of CLV models train and prediction process
    newcomers = TrainLSTMNewComers(
        date=date,
        time_indicator=time_indicator,
        order_count=order_count,
        data_source=data_source,
        data_query_path=data_query_path,
        time_period=time_period,
        directory=export_path,
        customer_indicator=customer_indicator,
        engaged_customers_results=engaged_customers_results,
        amount_indicator=amount_indicator)

    if job == 'train':
        newcomers.train_execute()
    if job == 'prediction':
        newcomers.prediction_execute()
    if job == 'train_prediction':
        newcomers.train_execute()
        newcomers.prediction_execute()

    newcomers_clv = newcomers.results
    del newcomers

    return {"next_purchase": {'data': raw_data},
            "purchase_amount": {'results': engaged_customers_results},
            "newcomers": {'results': newcomers_clv}}


if __name__ == '__main__':
    parser = argparse.ArgumentParser()
    parser.add_argument("-J", "--job", type=str,
                        help="""
                                train, prediction
                        """,
                        )
    parser.add_argument("-OC", "--order_count", type=str,
                        help="""
                                number of previous order count. It has been using for prediction next order frequency
                        """,
                        )
    parser.add_argument("-CI", "--customer_indicator", type=str,
                        help="""identifier of the customer (id)

                        """,
                        )
    parser.add_argument("-AI", "--amount_indicator", type=str,
                        help="""
                                data column n-amount (preferred numeric)
                        """,
                        )
    parser.add_argument("-DS", "--data_source", type=str,
                        help="""
                        AWS RedShift, BigQuery, PostgreSQL, csv, json files can be connected to system
                        """,
                        required=True)
    parser.add_argument("-DQP", "--data_query_path", type=str,
                        help="""
                        if there is file for data importing;
                            must be the path (e.g /../.../ab_test_raw_data.csv)
                        if there is ac- connection such as PostgreSQL / BigQuery
                            query must at the format "SELECT++++*+++FROM++ab_test_table_+++"
                        """,
                        required=True)
    parser.add_argument("-TI", "--time_indicator", type=str,
                        help="""
                        This can only be applied with date. It can be week, week_part, quarter, year, month.
                        Individually time indicator checks the date part is significantly
                        a individual group for data set or not.
                        If it is uses time_indicator as a  group
                        """,
                        )
    parser.add_argument("-TP", "--time_period", type=str,
                        help="""
                        This shows us to the time period
                        """,
                        )
    parser.add_argument("-EP", "--export_path", type=str,
                        help="""
                        Exporting path of the results set. Csv file of exporting.
                        """,
                        )
    arguments = parser.parse_args()
    args = {'job': arguments.job, 'order_count': arguments.order_count,
            'customer_indicator': arguments.customer_indicator,
            'amount_indicator': arguments.amount_indicator,
            'data_source': arguments.data_source,
            'data_query_path': arguments.data_query_path,
            'time_period': arguments.time_period,
            'time_indicator': arguments.time_indicator,
            'export_path': arguments.export_path}
    print(args)
    main(**args)



import yaml
from os.path import join, abspath


def read_config(directory):
    with open(join(directory, "docs", "configs.yaml")) as file:
        config = yaml.full_load(file)
    return config


def read_params(directory):
    with open(join(directory, "docs", "test_parameters.yaml")) as file:
        config = yaml.full_load(file)
    return config


def conf(var):
    config = read_config(abspath(__file__).split("configs.py")[0])
    return {
             'config': {c: config[c] for c in config
                        if c not in
                        ['data_main_path', 'model_main_path', 'log_main_path', 'docs_main_path', 'folder_name']}
    }[var]


def hyper_conf(var):
    config = read_params(abspath(__file__).split("configs.py")[0])
    return {'purchase_amount_hyper': config['models']['purchase_amount']['hyper_params'],
            'next_purchase_hyper': config['models']['next_purchase']['hyper_params'],
            'newcomers_hyper': config['models']['newcomers']['hyper_params'],
            'purchase_amount': config['models']['purchase_amount']['params'],
            'next_purchase': config['models']['next_purchase']['params'],
            'newcomers': config['models']['newcomers']['params']
            }[var]


iteration = 30
boostrap_ratio = 0.5
weekdays = ['Mondays', 'Tuesdays', 'Wednesdays', 'Thursdays', 'Fridays', 'Saturdays', 'Sundays']
web_port_default = 7002
accepted_ratio_of_actual_order = 0.75
accept_threshold_for_loss_diff = 0.05
parameter_tuning_trials = 10


import warnings
from pandas import DataFrame, concat
import os
import shutil
from itertools import product
import subprocess
from psutil import virtual_memory


from tensorflow.keras.layers import Dense, LSTM, Input, BatchNormalization
from tensorflow.keras.optimizers import Adam
from tensorflow.keras.initializers import Ones
from tensorflow.keras.models import Model
from tensorflow.keras.models import model_from_json
from kerastuner.tuners import RandomSearch
from kerastuner.engine.hyperparameters import HyperParameters

try:
    from functions import *
    from configs import hyper_conf, accept_threshold_for_loss_diff, parameter_tuning_trials
    from data_access import *
    from utils import abspath_for_sample_data
except Exception as e:
    from .functions import *
    from .configs import hyper_conf, accept_threshold_for_loss_diff, parameter_tuning_trials
    from .data_access import *
    from .utils import abspath_for_sample_data


def model_from_to_json(path=None, weights_path=None, model=None, is_writing=False, lr=None):
    if is_writing:
        model_json = model.to_json()
        with open(path, "w") as json_file:
            json_file.write(model_json)
        model.save_weights(weights_path)
    else:
        json_file = open(path, 'r')
        loaded_model_json = json_file.read()
        json_file.close()
        model = model_from_json(loaded_model_json)
        try:
            model.load_weights(weights_path)
        except Exception as e:
            model.load_weights(weights_path)
            model.compile(loss='mae', optimizer=Adam(lr=lr), metrics=['mae'])
        return model


def updating_hyper_parameters_related_to_data():
    return None


def get_params(params, comb):
    count = 0
    for p in params:
        _p = type(params[p])(comb[count])
        params[p] = _p
        count += 1
    return params


class TrainLSTM:
    def __init__(self,
                 date=None,
                 time_indicator=None,
                 order_count=None,
                 data_source=None,
                 data_query_path=None,
                 time_period=None,
                 directory=None,
                 customer_indicator=None,
                 amount_indicator=None):
        self.directory = directory
        self.customer_indicator = customer_indicator
        self.time_indicator = time_indicator
        self.amount_indicator = amount_indicator
        self.params = hyper_conf('next_purchase') \
            if check_for_existing_parameters(self.directory,'next_purchase') is None else \
            check_for_existing_parameters(self.directory, 'next_purchase')
        self.hyper_params = get_tuning_params(hyper_conf('next_purchase_hyper'), self.params)
        self.optimized_parameters = {}
        self._p = None
        self.order_count = order_count
        self.time_period = time_period
        self.data, self.features, self.c_min_max, self.params = data_manipulation_np(date=date,
                                                                                     feature=amount_indicator,
                                                                                     time_indicator=time_indicator,
                                                                                     data_source=data_source,
                                                                                     data_query_path=data_query_path,
                                                                                     customer_indicator=customer_indicator,
                                                                                     params=self.params,
                                                                                     time_period=time_period,
                                                                                     directory=directory)

        self.hp = HyperParameters()
        self.model_data = {}
        self.input, self.model = None, None
        self.prev_model_date = check_model_exists(self.directory, "trained_next_purchase_model", self.time_period)
        self.result = None, None
        self.residuals, self.anomaly = [], []
        self.customers = list(self.data[customer_indicator].unique())
        self.results = DataFrame()
        self.get_actual_value = lambda _min, _max, _value: ((_max - _min) * _value) + _min
        self.max_date = max(self.data[self.time_indicator])
        self.future_date = self.max_date + datetime.timedelta(days=convert_time_preiod_to_days(self.time_period))
        self.model_data = {"x_train": None, "y_train": None, "x_test": None, "y_test": None}
        self.client_sample_sizes = []
        self.optimum_batch_size = 32

    def get_model_data(self, customer):
        _data = self.data[self.data[self.customer_indicator] == customer]
        try:
            data = arrange__data_for_model(df=_data, f=[self.features], parameters=self.params)
            if data['x_train'].shape[0] != 0:
                model_data[customer] = {}
                model_data[customer] = data
        except Exception as e:
            print(e)

    def data_preparation(self):
        global model_data
        model_data = {}
        print("number of customers :", len(self.customers))
        execute_parallel_run(self.customers, self.get_model_data, arguments=None, parallel=16)
        try_count = 0
        while try_count < 20:
            try:
                self.model_data = {"x_train": None, "y_train": None, "x_test": None, "y_test": None}
                self.client_sample_sizes = []
                for c in model_data:
                    if self.model_data['x_train'] is not None:
                        for _split in ['x_train', 'y_train', 'x_test', 'y_test']:
                            self.model_data[_split] = np.concatenate([self.model_data[_split], model_data[c][_split]])
                        self.client_sample_sizes.append(model_data[c]['x_train'].shape[0])
                    else:
                        self.client_sample_sizes.append(model_data[c]['x_train'].shape[0])
                        self.model_data = model_data[c]
                try_count = 20
            except Exception as e:
                print(e)
                try_count += 1

    def build_parameter_tuning_model(self, hp):
        self.input = Input(shape=(self.model_data['x_train'].shape[1], 1))
        lstm = LSTM(int(hp.Choice('units', self.hyper_params['units'])),
                    bias_initializer=Ones(),
                    kernel_initializer=Ones(),
                    use_bias=False,
                    activation=hp.Choice('activation', self.hyper_params['activation']),
                    dropout=0.1
                    )(self.input)
        lstm = BatchNormalization()(lstm)
        lstm = Dense(1)(lstm)
        model = Model(inputs=self.input, outputs=lstm)
        model.compile(loss='mae',
                      optimizer=Adam(lr=hp.Choice('lr', self.hyper_params['lr'])),
                      metrics=['mae'])
        return model

    def build_model(self):
        self.input = Input(shape=(self.model_data['x_train'].shape[1], 1))
        # LSTM layer
        lstm = LSTM(self.params['units'],
                    batch_size=self.params['batch_size'],
                    bias_initializer=Ones(),
                    kernel_initializer=Ones(),
                    use_bias=False,
                    recurrent_activation=self.params['activation'],
                    dropout=0.1
                    )(self.input)
        lstm = BatchNormalization()(lstm)
        lstm = Dense(1)(lstm)
        self.model = Model(inputs=self.input, outputs=lstm)
        self.model.compile(loss='mae', optimizer=Adam(lr=self.params['lr']), metrics=['mae'])

    def learning_process(self, save_model=True, history=False):
        if history:
            history = self.model.fit(self.model_data['x_train'],
                                     self.model_data['y_train'],
                                     batch_size=self.params['batch_size'],
                                     epochs=int(self.params['epochs']),
                                     verbose=1,
                                     validation_data=(self.model_data['x_test'], self.model_data['y_test']),
                                     shuffle=True)
        else:
            print("*"*5, "Fit Next Purchase Model", "*"*5)
            self.model.fit(self.model_data['x_train'],
                           self.model_data['y_train'],
                           batch_size=self.params['batch_size'],
                           epochs=int(self.params['epochs']),
                           verbose=1,
                           validation_data=(self.model_data['x_test'], self.model_data['y_test']),
                           shuffle=True)
        if save_model:
            model_from_to_json(path=model_path(self.directory,
                                               "trained_next_purchase_model",
                                               get_current_day(),
                                               self.time_period),
                               weights_path=weights_path(self.directory,
                                                         "trained_next_purchase_model",
                                                         get_current_day(),
                                                         self.time_period),
                               model=self.model,
                               is_writing=True)

        if history:
            return history

    def train_execute(self):
        print("*"*5, "Next purchase train model process ", "*"*5)
        if self.prev_model_date is None:
            self.data_preparation()
            self.parameter_tuning()
            self.build_model()
            self.learning_process()
        else:
            self.model = model_from_to_json(path=model_path(self.directory,
                                                            "trained_next_purchase_model",
                                                            self.prev_model_date,
                                                            self.time_period),
                                            weights_path=weights_path(self.directory,
                                                                      "trained_next_purchase_model",
                                                                      self.prev_model_date,
                                                                      self.time_period), lr=self.params['lr'])
            print("Previous model already exits in the given directory  '" + self.directory + "'.")

    def create_prediction_data(self, customers, _directory):
        """
        import data with .csv format to the 'temp_next_purchase_inputs'. with given batch.
        Example of directory and data:
            data:
                customer_indicator | amount_indicator | time_indicator
                ------------------------------------------------------
                   cust_1          |   20.3           |  2021-06-05
                   .......         |   .....          |   .....
                   cust_2          |   30.3           |  2021-06-08
                   .......         |   .....          |   .....
                   cust_3          |   30.3           |  2021-06-10


            directory : <self.directory>/temp_next_purchase_inputs/prediction_inputs_cust_1.csv
        """
        self.temp_data = self.data[self.data[self.customer_indicator].isin(customers)]
        self.temp_data.to_csv(_directory, index=False)

    def parallel_prediction(self):
        """
        *** Parallel Prediction Process ***
            1. Create input folder: 'temp_next_purchase_inputs'
            2. Create result folder for the prediction: 'temp_next_purchase_results'
            3. split users (split number of user count related customer_indicator parameter) and run each batch sequentially.
            3. Parallel process is running on 'next_purchase_prediction.py'. Check related .py file for details.
        """
        try: # create input folder
            os.mkdir(join(self.directory, "temp_next_purchase_results"))
        except Exception as e:
            print(e)
            print("recreating 'temp_next_purchase_results' folder ...")
            shutil.rmtree(join(self.directory, "temp_next_purchase_results", ""))
            os.mkdir(join(self.directory, "temp_next_purchase_results"))


        try: # create output folder for the results
            os.mkdir(join(self.directory, "temp_next_purchase_inputs"))
        except Exception as e:
            print(e)
            print("recreating 'temp_next_purchase_inputs' folder ...")
            shutil.rmtree(join(self.directory, "temp_next_purchase_inputs", ""))
            os.mkdir(join(self.directory, "temp_next_purchase_inputs"))
        # check cpu count and available memory in order to optimize batch sizes
        cpus = cpu_count() * int((virtual_memory().total / 1000000000) * 4)
        communicates = [i for i in range(cpus) if i % cpu_count() == 0] + [cpus - 1]
        _sample_size = int(len(self.customers) / cpus) + 1
        _sub_processes = []
        for i in range(cpus):
            print("main iteration :", str(i), " / ", str(cpus))
            _sample_customers = get_iter_sample(self.customers, i, cpus, _sample_size)
            if len(_sample_customers) != 0:
                _directory = join(self.directory,
                                  "temp_next_purchase_inputs",
                                  "prediction_inputs_" + _sample_customers[0] + ".csv")
                self.create_prediction_data(_sample_customers, _directory)
                cmd = """python {0}/next_purchase_prediction.py -P {1} -MD {2} -FD {3} -TP {4} -D {5} -IND {6}
                """.format(abspath_for_sample_data(), _directory, str(self.max_date)[0:10],
                           str(self.future_date)[0:10], self.time_period, self.directory,
                           "*".join([self.customer_indicator, self.amount_indicator, self.time_indicator]))
                _p = subprocess.Popen(cmd, shell=True)
                _sub_processes.append(_p)
                if i != 0 and i in communicates:
                    [p.communicate() for p in _sub_processes]
                    print("done!")
                    _sub_processes = []

    def create_prediction_result_data(self):
        """
        after the parallel process prediction, results are stored in the folder 'temp_next_purchase_results'.
        Now, it is time to merge them
        """
        print("merge predicted data ...")
        _import_files = glob.glob(join(self.directory, "temp_next_purchase_results", "*.csv"))
        li = []
        for f in _import_files: # merge all result files
            try:
                _df = pd.read_csv(f, index_col=None, header=0)
                li.append(_df)
            except Exception as e:
                print(e)
        self.results = concat(li, axis=0, ignore_index=True)
        self.results[self.time_indicator] = self.results[self.time_indicator].apply(lambda x: convert_str_to_day(x))
        # remove temp folders
        shutil.rmtree(join(self.directory, "temp_next_purchase_results", ""))
        shutil.rmtree(join(self.directory, "temp_next_purchase_inputs", ""))
        # filter the dates related to selected time period
        self.results = self.results[(self.results[self.time_indicator] > self.max_date) &
                                    (self.results[self.time_indicator] < self.future_date)]
        print(self.results.head())
        print("number of predicted customers :", len(self.results[self.customer_indicator].unique()))

    def prediction_execute(self):
        """
        This process is running sequentially for each customer and predicts next order of timestamp
        """
        print("*"*5, "PREDICTION", 5*"*")
        print("number of users :", len(self.customers))
        if self.model is None:
            _model_date = self.prev_model_date if self.prev_model_date is not None else get_current_day()
            # import trained model from temp. directory folder
            self.model = model_from_to_json(path=model_path(self.directory,
                                                            "trained_next_purchase_model",
                                                            _model_date,
                                                            self.time_period),
                                            weights_path=weights_path(self.directory,
                                                                      "trained_next_purchase_model",
                                                                      _model_date,
                                                                      self.time_period), lr=self.params['lr'])
        self.results = self.data[[self.time_indicator, 'time_diff', 'time_diff_norm', self.customer_indicator]]
        self.parallel_prediction() # this process will be triggered via next_purchase_prediction.py
        self.create_prediction_result_data() # merge all result data

    def initialize_keras_tuner(self):
        """
        Parameter tuning process is triggered via Keras-Turner Library.
        However, batch_size and epoch parameters of optimization are created individually.
        """

        self.hyper_params, self.optimum_batch_size = batch_size_hp_ranges(client_sample_sizes=self.client_sample_sizes,
                                                                          num_of_customers=len(self.customers),
                                                                          hyper_params=self.hyper_params)
        kwargs = {'directory': self.directory}
        tuner = RandomSearch(
            self.build_parameter_tuning_model,
            max_trials=parameter_tuning_trials,
            hyperparameters=self.hp,
            allow_new_entries=True,
            objective='loss', **kwargs)
        tuner.search(x=self.model_data['x_train'],
                     y=self.model_data['y_train'],
                     epochs=5,
                     batch_size=self.optimum_batch_size,
                     verbose=1,
                     validation_data=(self.model_data['x_test'], self.model_data['y_test']))
        for p in tuner.get_best_hyperparameters()[0].values:
            if p in list(self.params.keys()):
                self.params[p] = tuner.get_best_hyperparameters()[0].values[p]

    def batch_size_and_epoch_tuning(self):
        """
        It finds the optimum batch_size and epoch.
        Each chosen epoch and batch_size of model created via model.
        Last epoch of loss is lower than 'accept_threshold_for_loss_diff' decision be finalized.
        Epoch will test ascending format (e.g. 4, 8, 16, ..., 1024) in order to minimize time consumption
        of both parameter tuning and model creation.
        Batch size will test descending format (e.g. 1024, 512, ...4) in order to minimize time consumption
        of both parameter tuning and model creation.
        """

        counter = 0
        optimum_epoch_process_done = False
        epoch_bs_combs = list(product(sorted(self.hyper_params['batch_sizes'], reverse=True),
                                      sorted(self.hyper_params['epochs'])
                                      ))
        loss_values = []
        while not optimum_epoch_process_done:
            self.params['epochs'] = int(epoch_bs_combs[counter][1])
            self.params['batch_size'] = int(epoch_bs_combs[counter][0])
            self.build_model()
            _history = self.learning_process(save_model=False, history=True)
            if _history.history['loss'][-1] < accept_threshold_for_loss_diff:
                optimum_epoch_process_done = True
            loss_values.append({"epochs": self.params['epochs'],
                                "batch_size": self.params['batch_size'],
                                "loss": _history.history['loss'][-1]})
            counter += 1
            if counter >= parameter_tuning_trials:
                optimum_epoch_process_done = True
                loss_values_df = pd.DataFrame(loss_values).sort_values(by='loss', ascending=True)
                self.params['epochs'] = list(loss_values_df['epochs'])[0]
                self.params['batch_size'] = list(loss_values_df['batch_size'])[0]

    def remove_keras_tuner_folder(self):
        """
        removing keras tuner file. while you need to update the parameters it will affect rerun the parameter tuning.
        It won`t start unless the folder has been removed.
        """

        try:
            shutil.rmtree(join(self.directory, "untitled_project"))
        except Exception as e:
            print(" Parameter Tuning Keras Turner dummy files have already removed!!")

    def update_tuned_parameter_file(self):
        """
        tuned parameters are stored at 'test_parameters.yaml.' in given 'export_path' argument.
        """

        # existing model for 'purchase_amount' that means p. tuning was applied for both models.
        if check_for_existing_parameters(self.directory, 'purchase_amount') is not None:
            _params = read_yaml(self.directory, "test_parameters.yaml")
            _params['next_purchase'] = self.params
        else:
            print("None of parameter tuning for both Model has been observed.")
            _params = {'next_purchase': self.params}
        write_yaml(self.directory, "test_parameters.yaml", _params, ignoring_aliases=True)

    def parameter_tuning(self):
        if check_for_existing_parameters(self.directory, 'next_purchase') is None:
            self.initialize_keras_tuner()
            self.batch_size_and_epoch_tuning()
            self.remove_keras_tuner_folder()
            self.update_tuned_parameter_file()
        else:
            self.params = check_for_existing_parameters(self.directory, 'next_purchase')








import dash
import dash_core_components as dcc
import dash_html_components as html
import plotly.graph_objs as go
import pandas as pd
import numpy as np
import datetime
import webbrowser
import argparse

try:
    from .data_access import GetData
    from .utils import convert_date
    from .functions import convert_time_preiod_to_days, get_results
except Exception as e:
    from data_access import GetData
    from utils import convert_date
    from functions import convert_time_preiod_to_days, get_results


def get_raw_data(time_indicator, amount_indicator, data_source, data_query_path):
    try:
        source = GetData(data_query_path=data_query_path,
                         data_source=data_source,
                         time_indicator=time_indicator,
                         feature=amount_indicator)
        source.query_data_source()
        source.convert_feature()
        data = source.data
    except Exception as e:
        data = pd.DataFrame()
        print("no data is available")
    return data


def get_filters(results, customer_indicator, amount_indicator):
    data_pv = results.groupby(customer_indicator).agg({amount_indicator: "sum"}).reset_index()
    top_100_customers = list(data_pv.sort_values(amount_indicator,ascending=False)[customer_indicator])[0:100]
    worst_100_customers = list(data_pv.sort_values(amount_indicator, ascending=True)[customer_indicator])[0:100]
    calculation_selection = ['sum', 'mean']
    filter_datas = [top_100_customers, worst_100_customers, calculation_selection]
    filter_ids = ['top_100_customers', 'worst_100_customers', 'calculation_type']
    filter_sizes = [100, 100, 20]
    multiple_selection = [False] * 3
    values = ['all', 'all', 'sum']
    filters = list(zip(filter_ids, filter_datas, filter_sizes, multiple_selection, values))
    return 3, filters, filter_ids, top_100_customers, worst_100_customers


def decide_time_period(date, time_period):
    if time_period == 'day':
        return datetime.datetime.strptime(str(date)[0:10], "%Y-%m-%d")
    if time_period in ['month', '2*month', 'quarter', '6*month']:
        return datetime.datetime.strptime(str(date)[0:7], "%Y-%m")
    if time_period == 'year':
        return datetime.datetime.strptime(str(date)[0:4], "%Y")
    if time_period == 'week':  # only shows mondays for weekly comparison
        week_day = datetime.datetime.strptime(str(date)[0:10], "%Y-%m-%d").isoweekday()
        return datetime.datetime.strptime(str(date)[0:10], "%Y-%m-%d") - datetime.timedelta(days=week_day-1)
    if time_period == '2*week':  # only shows mondays for weekly comparison
        date = datetime.datetime.strptime(str(date)[0:10], "%Y-%m-%d")
        week_day = date.isoweekday()
        week = date.isocalendar()[1]
        mondays_cal_day_count = week_day-1 if week % 2 == 0 else week_day - 1 + 7
        return datetime.datetime.strptime(str(date)[0:10], "%Y-%m-%d") - datetime.timedelta(days=mondays_cal_day_count)


def get_data_time_period_column(data, results, time_indicator, time_period):
    data['data_type'] = 'actual'
    results['data_type'] = 'prediction'
    data = pd.concat([data, results])
    data = data[data[time_indicator] == data[time_indicator]]
    data[time_indicator + '_per_' + time_period] = data[time_indicator].apply(
        lambda x: decide_time_period(x, time_period))
    return data


def get_new_comer_list(row, customer_list, time_p_ind, customer_indicator):
    prev_customers = list(customer_list[customer_list[time_p_ind] < row[time_p_ind]][customer_indicator])
    if len(prev_customers) != 0:
        prev_customers = np.concatenate(np.array(prev_customers)).tolist()
        diffs = list(set(row[customer_indicator]) - set(prev_customers))
        return pd.Series([diffs, len(diffs) / (len(prev_customers) + len(diffs))])
    else:
        return pd.Series([row[customer_indicator], 1])


def get_churn_list(row, customer_list, time_p_ind, customer_indicator, max_time_indicator):
    prediction_customers = list(customer_list[customer_list[time_p_ind] == max_time_indicator][customer_indicator])
    if len(row[customer_indicator]) != 0 and len(prediction_customers) != 0:
        prediction_customers = np.concatenate(np.array(prediction_customers)).tolist()
        diffs = list(set(row[customer_indicator]) - set(prediction_customers))
        return pd.Series([diffs, len(diffs) / len(set(prediction_customers +row[customer_indicator]))])
    else:
        return pd.Series([[], 0])


def get_churn_and_new_comer_columns(data, time_p_ind, amount_indicator, customer_indicator):
    data_pv = data.groupby(time_p_ind).agg(
        {amount_indicator + "_sum": "sum",
         amount_indicator + "_mean": "mean",
         customer_indicator: lambda x: list(np.unique(x))}).reset_index()

    data_pv = data_pv.sort_values(by=time_p_ind, ascending=True)
    customer_list = data_pv[[time_p_ind, customer_indicator]]
    predicted_time_period = max(customer_list[time_p_ind])
    data_pv[['new_comer_list', 'new_comer_ratio']] = data_pv.apply(
        lambda row: get_new_comer_list(row, customer_list, time_p_ind, customer_indicator), axis=1)
    data_pv[['churn_list', 'churn_ratio']] = data_pv.apply(
        lambda row: get_churn_list(row, customer_list, time_p_ind, customer_indicator, predicted_time_period), axis=1)
    data = pd.merge(data, data_pv[[time_p_ind, 'new_comer_list', 'new_comer_ratio']], on=time_p_ind, how='left')
    data = pd.merge(data, data_pv[[time_p_ind, 'churn_list', 'churn_ratio']], on=time_p_ind, how='left')
    return data, max(customer_list[time_p_ind])


def get_new_comer_churn_data(data, customer_indicator, amount_indicator, time_p_ind):
    data['is_new_comer'] = data.apply(lambda row: True if row[customer_indicator] in row['new_comer_list'] else False,
                                      axis=1)
    data['is_churn'] = data.apply(lambda row: True if row[customer_indicator] in row['churn_list'] else False, axis=1)
    new_comer_data = data.query("is_new_comer == True").groupby(time_p_ind).agg(
        {amount_indicator + "_sum": "sum",
         amount_indicator + "_mean": "mean", 'new_comer_ratio': 'first'}).reset_index()
    churn_data = data.query("is_churn == True").groupby(time_p_ind).agg(
        {amount_indicator + "_sum": "sum",
         amount_indicator + "_mean": "mean", 'churn_ratio': 'first'}).reset_index()
    return new_comer_data, churn_data


def pivoting_data_per_time_period(data,
                                  results,
                                  time_indicator,
                                  time_period,
                                  amount_indicator,
                                  customer_indicator,
                                  top_100_customers,
                                  worst_100_customers):
    time_p_ind = time_indicator + '_per_' + time_period
    data = get_data_time_period_column(data, results, time_indicator, time_period)
    data[amount_indicator + "_sum"], data[amount_indicator + "_mean"] = data[amount_indicator], data[amount_indicator]
    data = data[data[amount_indicator] == data[amount_indicator]]
    data_pv = data.groupby([time_p_ind, 'data_type']).agg(
        {amount_indicator + "_sum": "sum",
         amount_indicator + "_mean": "mean",
         customer_indicator: lambda x: list(np.unique(x))}).reset_index()
    data, predicted_time_period = get_churn_and_new_comer_columns(data, time_p_ind, amount_indicator, customer_indicator)
    new_comer_data, churn_data = get_new_comer_churn_data(data, customer_indicator, amount_indicator, time_p_ind)
    top_100_data = data[data[customer_indicator].isin(top_100_customers)]
    worst_100_data = data[data[customer_indicator].isin(worst_100_customers)]
    return data_pv.query("data_type == 'actual'"), data_pv.query("data_type == 'prediction'"), \
           top_100_data, worst_100_data, predicted_time_period, new_comer_data, churn_data


def adding_filter_to_pane(added_filters, f_style):
    return html.Div(added_filters, style=f_style)


def adding_plots_to_pane(plot_id, hover_data, size):
    return html.Div([
        dcc.Graph(
            id=plot_id,
            hoverData={'points': [hover_data]}
        )
    ], style={'width': str(size) + '%', 'display': 'inline-block', 'padding': '0 90'})


def adding_filter(filter_id, labels, size, is_multi_select, value):
    return html.Div([
        html.Div(filter_id, style={'width': '40%', 'float': 'left', 'display': 'inline-block'}),
        dcc.Dropdown(
            id=filter_id,
            options=[{'label': i, 'value': i} for i in labels],
            multi=True if is_multi_select else False,
            value=value
        )
    ],
        style={'width': str(size) + '%', 'display': 'inline-block'})


"""
Time Line Chart with filter selected average/Total value per time period with predicted life time value
Top 100 customer of Time Period Change
Worst 100 Customer Of Time Period Value Change
New Comers of Time line with selected date from Time Line Chart
Churn Customers of Time line with selected date from Time Line Chart
"""


def get_hover_data(data, time_indicator, time_period, number_of_graph):
    value = sorted(list(data[time_indicator + '_per_' + time_period].unique()))
    return [{'customdata': str(value[-min(3, len(value))])}] * number_of_graph


def create_dashboard(customer_indicator, amount_indicator, directory,
                     time_indicator, time_period, data_query_path, data_source):
    external_stylesheets = ['https://codepen.io/chriddyp/pen/bWLwgP.css']
    app = dash.Dash(__name__, external_stylesheets=external_stylesheets)
    try:
        data = get_raw_data(time_indicator, amount_indicator, data_source, data_query_path)
        results = get_results(directory, time_period)
        results = results[results[amount_indicator] == results[amount_indicator]]
        num_f_p, filters, filter_ids, top_100_customers, worst_100_customers = get_filters(results,
                                                                                           customer_indicator,
                                                                                           amount_indicator)
        app.layout = html.Div()
        if len(data) == 0:
            return app
    except Exception as e:
        print(e)
        app.layout = html.Div()
        return app
    actual, prediction, top_100_data, worst_100_data, predicted_time_period, \
    new_comer_data, churn_data = pivoting_data_per_time_period(
                                                                data,
                                                                results,
                                                                time_indicator,
                                                                time_period,
                                                                amount_indicator,
                                                                customer_indicator,
                                                                top_100_customers,
                                                                worst_100_customers)

    filter_style = {
                    'borderBottom': 'thin lightgrey solid',
                    'backgroundColor': 'rgb(250, 250, 250)',
                    'padding': '10px 5px'
    }
    plot_ids = ['time_line_of_values', 'top_100_customer', 'worst_100_customer',
                'churn_customer', 'new_comer_customer', 'churn_rate', 'new_comer_engage_rate']
    plot_sizes = [99, 24, 24, 24, 24, 24, 24]
    hover_datas = get_hover_data(actual, time_indicator, time_period, len(plot_ids))
    plots = list(zip(plot_ids, plot_sizes, hover_datas))
    # adding filters
    pane_count = int(len(filters) / num_f_p) if int(len(filters) / num_f_p) == len(filters) / num_f_p else int(
        len(filters) / num_f_p) + 1
    components = []
    for i in range(pane_count):
        _filters = filters[i * num_f_p:(i + 1) * num_f_p] if i != pane_count - 1 else filters[i * num_f_p:]
        _pane = adding_filter_to_pane([adding_filter(f[0], f[1], f[2], f[3], f[4]) for f in _filters], filter_style)
        components.append(_pane)
    # adding plots
    for p in plots:
        components.append(adding_plots_to_pane(p[0], p[2], p[1]))
    app.layout = html.Div(components)
    # Churn Rate
    """
    HoverData from clv time line selected date of customers who are newcomer of time line till the predicted time period
    """
    @app.callback(
        dash.dependencies.Output(plot_ids[5], 'figure'),
        [dash.dependencies.Input('time_line_of_values', 'hoverData')]
    )
    def update_graph_4(hover_data):
        starting_date = hover_data['points'][0]['customdata']
        dff = churn_data[churn_data[time_indicator + '_per_' + time_period] == starting_date]
        if len(dff) == 0:
            return {"data": [], "layout": go.Layout(height=600, title="Churn Customers Of Purchase Time Line")}
        else:
            churn_rate = \
            list(churn_data[churn_data[time_indicator + '_per_' + time_period] == starting_date]['churn_ratio'])[0]
            return_rate = 1 - churn_rate
            trace = [go.Pie(labels=['Churn Rate', 'Return Rate'],
                            values=[churn_rate, return_rate],
                            )
                     ]
            return {"data": trace,
                    "layout": go.Layout(height=600, title=" Churn Rate (%) per " + time_period)}

    # New Comer Ratio
    """
    HoverData from clv time line selected date of customers who are newcomer of time line till the predicted time period
    """
    @app.callback(
        dash.dependencies.Output(plot_ids[6], 'figure'),
        [dash.dependencies.Input('time_line_of_values', 'hoverData')]
    )
    def update_graph_4(hover_data):
        starting_date = hover_data['points'][0]['customdata']
        dff = new_comer_data[new_comer_data[time_indicator + '_per_' + time_period] == starting_date]
        if len(dff) == 0:
            return {"data": [], "layout": go.Layout(height=600, title="Churn Customers Of Purchase Time Line")}
        else:
            newc_rate = \
            list(churn_data[churn_data[time_indicator + '_per_' + time_period] == starting_date]['churn_ratio'])[0]
            engaged_users = 1 - newc_rate
            trace = [go.Pie(labels=['Newcomer User Ratio', 'Engaged User Ratio'],
                            values=[newc_rate, engaged_users],
                            )
                     ]
            return {"data": trace,
                    "layout": go.Layout(height=600, title=" Newcomer Rate (%) per " + time_period)}

    # Churn Customers Of Purchase Time Line
    """
    HoverData from clv time line selected date of customers who are newcomer of time line till the predicted time period
    """
    @app.callback(
        dash.dependencies.Output(plot_ids[3], 'figure'),
        [dash.dependencies.Input(f, 'value') for f in [filter_ids[-1]]] +
        [dash.dependencies.Input('time_line_of_values', 'hoverData')]
    )
    def update_graph_4(sum_or_avg, hover_data):
        starting_date = hover_data['points'][0]['customdata']
        value_column = amount_indicator + "_sum" if sum_or_avg == 'sum' else amount_indicator + "_mean"
        dff = churn_data[churn_data[time_indicator + '_per_' + time_period] >= starting_date]
        func = {value_column: "sum"} if sum_or_avg == 'sum' else {amount_indicator + "_mean": "mean"}
        dff = dff.groupby(time_indicator + '_per_' + time_period).agg(func).reset_index()
        if len(dff) == 0:
            return {"data": [], "layout": go.Layout(height=600, title="Churn Customers Of Purchase Time Line")}
        else:
            trace = [go.Scatter(x=dff[time_indicator + '_per_' + time_period],
                                y=dff[value_column],
                                mode='markers+lines',
                                name=value_column)
                     ]
            return {"data": trace,
                    "layout": go.Layout(height=600, title=" Churn Customers Of Purchase Time Line")}

    # top 100 customers
    @app.callback(
        dash.dependencies.Output(plot_ids[1], 'figure'),
        [dash.dependencies.Input(f, 'value') for f in [filter_ids[0], filter_ids[-1]]]
    )
    def update_graph_1(_customer, sum_or_avg):
        dff = top_100_data[top_100_data[customer_indicator] == _customer] if _customer != 'all' else top_100_data
        value_column = amount_indicator + "_sum" if sum_or_avg == 'sum' else amount_indicator + "_mean"
        func = {value_column: "sum"} if sum_or_avg == 'sum' else {amount_indicator + "_mean": "mean"}
        dff = dff.groupby(time_indicator + '_per_' + time_period).agg(func).reset_index()
        title = "Top 100 the most Engaged Customers of Average Values per " + time_period
        if sum_or_avg == 'sum':
            title = title.replace("Average", "Sum")
        if len(dff) == 0:
            return {"data": [], "layout": go.Layout(height=600, title=title)}
        else:
            trace = go.Scatter(x=dff[time_indicator + '_per_' + time_period],
                               y=dff[value_column],
                               mode='markers+lines',
                               name=value_column)
            return {"data": [trace],
                    "layout": go.Layout(height=600, title=title)}

    # worst 100 customers
    @app.callback(
        dash.dependencies.Output(plot_ids[2], 'figure'),
        [dash.dependencies.Input(f, 'value') for f in [filter_ids[1], filter_ids[-1]]]
    )
    def update_graph_2(_customer, sum_or_avg):
        dff = worst_100_data[worst_100_data[customer_indicator] == _customer] if _customer != 'all' else worst_100_data
        value_column = amount_indicator + "_sum" if sum_or_avg == 'sum' else amount_indicator + "_mean"
        func = {value_column: "sum"} if sum_or_avg == 'sum' else {amount_indicator + "_mean": "mean"}
        dff = dff.groupby(time_indicator + '_per_' + time_period).agg(func).reset_index()
        title = "Top 100 the least Engaged Customers of Average Values per " + time_period
        if sum_or_avg == 'sum':
            title = title.replace("Average", "Sum")
        if len(dff) == 0:
            return {"data": [], "layout": go.Layout(height=600, title=title)}
        else:
            trace = [go.Scatter(x=dff[time_indicator + '_per_' + time_period],
                                y=dff[value_column],
                                mode='markers+lines',
                                name=value_column)]
            return {"data": trace,
                    "layout": go.Layout(height=600, title=title)}

    # Time Line Of CLV Prediction
    """
    HoverData from clv time line selected date of customers who are newcomer of time line till the predicted time period
    """
    @app.callback(
        dash.dependencies.Output(plot_ids[0], 'figure'),
        [dash.dependencies.Input(f, 'value') for f in [filter_ids[-1]]]
    )
    def update_graph_4(sum_or_avg):
        value_column = amount_indicator + "_sum" if sum_or_avg == 'sum' else amount_indicator + "_mean"
        title = "SUM of " if sum_or_avg == 'sum' else "Average of "
        title += "CLV Prediction per " + time_period
        _prev_time_period = predicted_time_period - datetime.timedelta(days=convert_time_preiod_to_days(time_period))

        if len(actual) == 0 and len(prediction) == 0:
            return {"data": [], "layout": go.Layout(height=600, title=title)}
        else:
            trace = [go.Scatter(x=actual[time_indicator + '_per_' + time_period],
                                y=actual[value_column],
                                mode='markers+lines',
                                customdata=actual[time_indicator + '_per_' + time_period],
                                name=value_column + " Actual"),
                     go.Scatter(x=prediction[time_indicator + '_per_' + time_period],
                                y=prediction[value_column],
                                mode='markers+lines',
                                name='CLV Prediction')]
            return {"data": trace,
                    "layout": go.Layout(height=600, title=title)}

    # Newcomer Customers Of Customer Value Prediction
    """
    HoverData from clv time line selected date of customers who are newcomer of time line till the predicted time period
    """
    @app.callback(
        dash.dependencies.Output(plot_ids[4], 'figure'),
        [dash.dependencies.Input(f, 'value') for f in [filter_ids[-1]]] +
        [dash.dependencies.Input('time_line_of_values', 'hoverData')]
    )
    def update_graph_4(sum_or_avg, hover_data):
        starting_date = hover_data['points'][0]['customdata'] # '2018-01-01 00:00:00'
        value_column = amount_indicator + "_sum" if sum_or_avg == 'sum' else amount_indicator + "_mean"
        dff = new_comer_data[new_comer_data[time_indicator + '_per_' + time_period] >= starting_date]
        if len(dff) == 0:
            return {"data": [], "layout": go.Layout(height=600,
                                                    title="Newcomer Customer Value Prediction")}
        else:
            trace = [go.Scatter(x=dff[time_indicator + '_per_' + time_period],
                                y=dff[value_column],
                                mode='markers+lines',
                                customdata=dff[time_indicator + '_per_' + time_period],
                                name=value_column)]
            return {"data": trace,
                    "layout": go.Layout(height=600, title="Newcomer Customer Value Prediction")}

    webbrowser.open('http://127.0.0.1:8050/')
    app.run_server(debug=False)


if __name__ == '__main__':
    parser = argparse.ArgumentParser()
    parser.add_argument("-CI", "--customer_indicator", type=str,
                        help="""identifier of the customer (id)
##
                        """,
                        )
    parser.add_argument("-AI", "--amount_indicator", type=str,
                        help="""
                                data column n-amount (preferred numeric)
                        """,
                        )
    parser.add_argument("-TI", "--time_indicator", type=str,
                        help="""
                        This can only be applied with date. It can be only day..
                        Individually time indicator checks the date part is significantly
                        a individual group for data set or not.
                        If it is uses time_indicator as a  group
                        """,
                        )
    parser.add_argument("-TP", "--time_period", type=str,
                        help="""
                        This shows us to the time period
                        """,
                        )
    parser.add_argument("-EP", "--export_path", type=str,
                        help="""
                        Exporting path of the results set. Csv file of exporting.
                        """,
                        )
    parser.add_argument("-DS", "--data_source", type=str,
                        help="""
                        AWS RedShift, BigQuery, PostgreSQL, csv, json files can be connected to system
                        """,
                        required=True)
    parser.add_argument("-DQP", "--data_query_path", type=str,
                        help="""
                        if there is file for data importing;
                            must be the path (e.g /../.../ab_test_raw_data.csv)
                        if there is ac- connection such as PostgreSQL / BigQuery
                            query must at the format "SELECT++++*+++FROM++ab_test_table_+++"
                        """,
                        required=True)
    arguments = parser.parse_args()
    args = {
            'customer_indicator': arguments.customer_indicator,
            'amount_indicator': arguments.amount_indicator,
            'directory': arguments.export_path,
            'time_period': arguments.time_period,
            'time_indicator': arguments.time_indicator,
            'data_source': arguments.time_period,
            'data_query_path': arguments.data_query_path}
    print(args)
    create_dashboard(**args)


import setuptools
from setuptools import find_packages

install_requires_list = [
        "psutil",
        "numpy >= 1.18.1",
        "pandas >= 0.25.3",
        "scipy >= 1.4.1 ",
        "PyYAML",
        "schedule >= 0.6.0",
        "multiprocess >= 0.70.9",
        "google-cloud-bigquery",
        "mysql-connector-python",
        "plotly >=  4.5.0",
        "dash-html-components >= 1.0.2",
        "dash-core-components >=  1.8.0",
        "dash >= 1.9.0",
        "threaded >= 4.0.8",
        "pytest-shutil >= 1.7.0",
        "python-dateutil >= 2.8.1",
        "random2 >= 1.0.1",
        "psycopg2-binary",
        "argparse",
        "python-math",
        "statsmodels >= 0.12.1",
        "keras-tuner >= 1.0.2",
        "multiprocess >= 0.70.9",
        "tensorflow >= 2.2.0",
        "Keras >= 2.3.1"
    ]


with open("README.md", "r") as fh:
    long_description = fh.read()


setuptools.setup(
    name="clv_prediction",
    version="0.1.8",
    author="Caglan Akpinar",
    author_email="cakpinar23@gmail.com",
    description="clv prediction applying with deep learning",
    long_description=long_description,
    long_description_content_type="text/markdown",
    keywords='CLV, Customer Lifetime Value, Lifetime Prediction',
    packages= find_packages(exclude='__pycache__'),
    py_modules=['clv', 'clv/docs'],
    install_requires=install_requires_list,
    url="https://github.com/caglanakpinar/clv_prediction",
    include_package_data=True,
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    python_requires='>=3',
)



import os, inspect
import yaml
import json
import datetime
from os.path import join, abspath
from multiprocessing import cpu_count
import threading
import subprocess


def read_yaml(directory, filename):
    with open(join(directory, "", filename)) as file:
        docs = yaml.full_load(file)
    return docs


def write_yaml(directory, filename, data, ignoring_aliases=False):
    if ignoring_aliases:
        yaml.Dumper.ignore_aliases = lambda *args : True

    with open(join(directory, "", filename), 'w') as file:
        if ignoring_aliases:
            yaml.dump(data, file, default_flow_style=False)
        else:
            yaml.dump(data, file)


def read_write_to_json(directory, filename, data, is_writing):
    if is_writing:
        with open(join(directory, "", filename), 'w') as file:
            json.dump(data, file)
    else:
        with open(join(directory, "", filename), "r") as file:
            data = json.loads(file.read())
        return data


def model_path(comb, group, model):
    return "_".join(["_".join([str(i[0]) + "*" + str(i[1]) for i in zip(group, comb)]), model]) + ".json"


def convert_date(date):
    if date not in ['', None]:
        if len(date) == 0:
            format_str = '%Y-%m-%d'
        if len(date) == 16:
            format_str = '%Y-%m-%d %H:%M'
        if len(date) > 16:
            format_str = '%Y-%m-%d %H:%M:%S.%f'
        if not str:
            date = datetime.datetime.strptime(date, format_str)
    else:
        date = datetime.datetime.now() + datetime.timedelta(minutes=2)
    return date


def get_current_day(replace=True):
    return str(datetime.datetime.now())[0:10].replace("-", "") if replace else str(datetime.datetime.now())[0:10]


def get_result_data_path(directory, time_period, max_date):
    return join(directory, "results_" + time_period + "_" +
                str(max_date)[0:10].replace("-", "") + "_" + get_current_day() + ".csv")


def convert_str_to_day(x):
    try:
        return datetime.datetime.strptime(str(x)[0:10], "%Y-%m-%d")
    except Exception as e:
        return None


def convert_feature(value):
    try:
        if value == value:
            return value
        else:
            return None
    except Exception as e:
        print(e)
        return None


def get_folder_path():
    return abspath(__file__).split("utils.py")[0]


def check_for_existing_parameters(directory, model):
    try:
        params = read_yaml(directory, "test_parameters.yaml")[model]
    except Exception as e:
        params = None
    return params


def get_iter_sample(s_values, i, iters, cpus):
    if i != iters - 1:
        return s_values[(i * cpus): ((i+1) * cpus)]
    else:
        return s_values[(i * cpus):]


def current_date_to_day():
    """
    recent date of converting to datetime from isoformat.
    :return: datetime
    """
    return datetime.datetime.strptime(str(datetime.datetime.now())[0:19], '%Y-%m-%d')


def execute_parallel_run(values, executor, parallel=2, arguments=None):
    """
    main concept of parallelize the processes. It checks the cpu count and multiplies with 4.
    On each iteration it triggers cpu * 4 threads sequentially.
    :param values: list of values in order to trigger thread.
    :param executor: function in order to call
    :param parallel: number of thread
    :param arguments: arguments for the function
    :return:
    """
    global process
    cpus = cpu_count()
    if parallel < cpus * 16:
        parallel = cpus * 16
    iters = int(len(values) / parallel) + 1
    print("number of iterations :", iters)
    for i in range(iters):
        if i % 10 == 0:
            print("iteration :", i)
        _sample_values = get_iter_sample(values, i, iters, parallel)
        for v in _sample_values:
            if arguments:
                process = threading.Thread(target=executor, args=(v, arguments, ))
            else:
                process = threading.Thread(target=executor, args=(v,))
            process.deamon = True
            process.start()
        process.join()
    del process
    return ""


def abspath_for_sample_data():
    """
    get customer_analytics path. Ex: ....../customer_analytics
    :return: current folder path
    """
    currentdir = os.path.dirname(os.path.abspath(inspect.getfile(inspect.currentframe())))
    base_name = os.path.basename(currentdir)
    while base_name != 'clv':
        currentdir = os.path.dirname(os.path.abspath(inspect.getfile(inspect.currentframe())))
        base_name = os.path.basename(currentdir)
    return currentdir


def sample_size_calculation():
    cpus = cpu_count()
    return 1250 * cpus









import warnings
import pandas as pd
import os
import shutil
from itertools import product
import glob
os.environ['TF_CPP_MIN_LOG_LEVEL'] = '4'

import tensorflow as tf
sess = tf.compat.v1.Session()

from tensorflow.keras.layers import Dense, LSTM, Input, BatchNormalization
from tensorflow.keras.optimizers import Adam
from tensorflow.keras.initializers import Ones
from tensorflow.keras.models import Model
from tensorflow.keras.models import model_from_json
from kerastuner.tuners import RandomSearch
from kerastuner.engine.hyperparameters import HyperParameters

try:
    from functions import *
    from configs import hyper_conf, accept_threshold_for_loss_diff, parameter_tuning_trials
    from data_access import *
except Exception as e:
    from .functions import *
    from .configs import hyper_conf, accept_threshold_for_loss_diff, parameter_tuning_trials
    from .data_access import *


def model_from_to_json(path=None, weights_path=None, model=None, is_writing=False):
    if is_writing:
        model_json = model.to_json()
        with open(path, "w") as json_file:
            json_file.write(model_json)
        model.save_weights(weights_path)
    else:
        json_file = open(path, 'r')
        loaded_model_json = json_file.read()
        json_file.close()
        model = model_from_json(loaded_model_json)
        model.load_weights(weights_path)
        return model


def updating_hyper_parameters_related_to_data():
    return None


def get_params(params, comb):
    count = 0
    for p in params:
        _p = type(params[p])(comb[count])
        params[p] = _p
        count += 1
    return params


class TrainLSTMNewComers:
    def __init__(self,
                 date=None,
                 time_indicator=None,
                 order_count=None,
                 data_source=None,
                 data_query_path=None,
                 time_period=None,
                 directory=None,
                 engaged_customers_results=pd.DataFrame(),
                 customer_indicator=None,
                 amount_indicator=None):
        self.sess = tf.compat.v1.Session()
        self.directory = directory
        self.customer_indicator = customer_indicator
        self.time_indicator = time_indicator
        self.amount_indicator = amount_indicator
        self.params = hyper_conf('newcomers') \
            if check_for_existing_parameters(self.directory,'newcomers') is None else \
            check_for_existing_parameters(self.directory, 'newcomers')
        self.hyper_params = get_tuning_params(hyper_conf('newcomers_hyper'), self.params)
        self.optimized_parameters = {}
        self._p = None
        self.order_count = order_count
        self.time_period = time_period
        self.engaged_customers_results = engaged_customers_results
        self.data, self.features, \
        self.average_amount, self.min_max = data_manipulation_nc(date=date,
                                                                 order_count=order_count,
                                                                 amount_indicator=amount_indicator,
                                                                 time_indicator=time_indicator,
                                                                 data_source=data_source,
                                                                 data_query_path=data_query_path,
                                                                 customer_indicator=customer_indicator,
                                                                 directory=directory)

        self.hp = HyperParameters()
        self.model_data = {}
        self.input, self.model = None, None
        self.prev_model_date = check_model_exists(self.directory, "trained_newcomers_model", self.time_period)
        self.residuals, self.anomaly = [], []
        self.results = pd.DataFrame()
        self.get_actual_value = lambda _min, _max, _value: ((_max - _min) * _value) + _min if _value >= 0 else _min
        self.max_date = max(self.data[self.time_indicator])
        self.future_date = self.max_date + datetime.timedelta(days=convert_time_preiod_to_days(self.time_period))
        self.model_data = {"x_train": None, "y_train": None, "x_test": None, "y_test": None}
        self.client_sample_sizes = []
        self.optimum_batch_size = 8

    def data_preparation(self):
        self.model_data = arrange__data_for_model(df=self.data, f=[self.features], parameters=self.params)

    def build_parameter_tuning_model(self, hp):
        self.input = Input(shape=(self.model_data['x_train'].shape[1], 1))
        lstm = LSTM(int(hp.Choice('units', self.hyper_params['units'])),
                    bias_initializer=Ones(),
                    kernel_initializer=Ones(),
                    use_bias=False,
                    activation=hp.Choice('activation', self.hyper_params['activation']),
                    dropout=0.1
                    )(self.input)
        lstm = BatchNormalization()(lstm)
        lstm = Dense(1)(lstm)
        model = Model(inputs=self.input, outputs=lstm)
        model.compile(loss='mae',
                      optimizer=Adam(lr=hp.Choice('lr', self.hyper_params['lr'])),
                      metrics=['mae'])
        return model

    def init_tf(self):
        self.sess.close()
        import tensorflow as tf
        self.sess = tf.compat.v1.Session()

        from tensorflow.keras.layers import Dense, LSTM, Input, BatchNormalization
        from tensorflow.keras.optimizers import Adam
        from tensorflow.keras.initializers import Ones
        from tensorflow.keras.models import Model
        from tensorflow.keras.models import model_from_json
        from kerastuner.tuners import RandomSearch
        from kerastuner.engine.hyperparameters import HyperParameters

    def build_model(self, prediction=False):
        if prediction:
            self.init_tf()
        self.input = Input(shape=(self.model_data['x_train'].shape[1], 1))
        # LSTM layer
        lstm = LSTM(self.params['units'],
                    batch_size=self.params['batch_size'],
                    bias_initializer=Ones(),
                    kernel_initializer=Ones(),
                    use_bias=False,
                    recurrent_activation=self.params['activation'],
                    dropout=0.1
                    )(self.input)
        lstm = BatchNormalization()(lstm)
        lstm = Dense(1)(lstm)
        self.model = Model(inputs=self.input, outputs=lstm)
        self.model.compile(loss='mae', optimizer=Adam(lr=self.params['lr']), metrics=['mae'])

    def learning_process(self, save_model=True, history=False, show_epochs=True):
        verbose = 1 if show_epochs else 0
        if history:
            history = self.model.fit(self.model_data['x_train'],
                                     self.model_data['y_train'],
                                     batch_size=self.params['batch_size'],
                                     epochs=int(self.params['epochs']),
                                     verbose=1,  # verbose = 1 if there if history = True
                                     validation_data=(self.model_data['x_test'], self.model_data['y_test']),
                                     shuffle=True)
        else:
            if save_model:
                print("*"*5, "Fit Newcomers CLV Model", "*"*5)
            self.model.fit(self.model_data['x_train'],
                           self.model_data['y_train'],
                           batch_size=self.params['batch_size'],
                           epochs=int(self.params['epochs']),
                           verbose=verbose,
                           validation_data=(self.model_data['x_test'], self.model_data['y_test']),
                           shuffle=True)
        if save_model:
            model_from_to_json(path=model_path(self.directory,
                                               "trained_newcomers_model",
                                               get_current_day(),
                                               self.time_period),
                               weights_path=weights_path(self.directory,
                                                         "trained_newcomers_model",
                                                         get_current_day(),
                                                         self.time_period),
                               model=self.model,
                               is_writing=True)

        if history:
            return history

    def train_execute(self):
        print("*"*5, "Newcomer CLV Prediction train model process ", "*"*5)
        if self.prev_model_date is None:
            self.data_preparation()
            self.parameter_tuning()
            self.build_model()
            self.learning_process()
        else:
            self.model = model_from_to_json(path=model_path(self.directory,
                                                            "trained_newcomers_model",
                                                            self.prev_model_date,
                                                            self.time_period),
                                            weights_path=weights_path(self.directory,
                                                                      "trained_newcomers_model",
                                                                      self.prev_model_date,
                                                                      self.time_period))
            print("Previous model already exits in arrange__data_for_model the given directory  '" + self.directory + "'.")

    def prediction_execute(self):
        print("*"*5, "PREDICTION", 5*"*")
        if self.model is None:
            _model_date = self.prev_model_date if self.prev_model_date is not None else get_current_day()
            self.model = model_from_to_json(path=model_path(self.directory,
                                                            "trained_newcomers_model",
                                                            _model_date,
                                                            self.time_period),
                                            weights_path=weights_path(self.directory,
                                                                      "trained_newcomers_model",
                                                                      _model_date,
                                                                      self.time_period))
        # daily calculations, day by day
        while self.max_date < self.future_date:
            print("date :", self.max_date)
            self.model_data = arrange__data_for_model(self.data, [self.features], self.params)
            self.build_model(prediction=True)
            self.learning_process(save_model=False, history=False, show_epochs=False)
            x = arrange__data_for_model(self.data, [self.features], self.params, is_prediction=True)
            _pred = pd.DataFrame([{self.time_indicator: self.max_date, "order_count": self.model.predict(x)[0][-1]}])
            self.data, self.results = pd.concat([self.data, _pred]), pd.concat([self.results, _pred])
            self.max_date += datetime.timedelta(days=1)
            del self.model_data, x, self.model
        for i in ['min_' + self.features, 'max_' + self.features]:
            self.results[i] = self.min_max[i]
        self.results[self.features] = self.results.apply(
            lambda row: self.get_actual_value(_min=row['min_' + self.features],
                                              _max=row['max_' + self.features],
                                              _value=row[self.features]), axis=1)
        self.results[self.amount_indicator] = self.results[self.features] * self.average_amount
        self.results[self.customer_indicator] = "newcomers"
        self.results['data_type'] = "prediction"
        self.results = self.results[['data_type', self.customer_indicator, self.time_indicator, self.amount_indicator]]
        print("result file : ", get_result_data_path(self.directory, self.time_period, self.max_date))
        pd.concat([self.results, self.engaged_customers_results]).to_csv(
            get_result_data_path(self.directory, self.time_period, self.max_date), index=False)

    def initialize_keras_tuner(self):
        """
        Parameter tuning process is triggered via Keras-Turner Library.
        However, batch_size and epoch parameters of optimization are created individually.
        """
        kwargs = {'directory': self.directory}
        tuner = RandomSearch(
            self.build_parameter_tuning_model,
            max_trials=parameter_tuning_trials,
            hyperparameters=self.hp,
            allow_new_entries=True,
            objective='loss', **kwargs)
        tuner.search(x=self.model_data['x_train'],
                     y=self.model_data['y_train'],
                     epochs=5,
                     batch_size=self.optimum_batch_size,
                     verbose=1,
                     validation_data=(self.model_data['x_test'], self.model_data['y_test']))
        for p in tuner.get_best_hyperparameters()[0].values:
            if p in list(self.params.keys()):
                self.params[p] = tuner.get_best_hyperparameters()[0].values[p]

    def batch_size_and_epoch_tuning(self):
        """
        It finds the optimum batch_size and epoch.
        Each chosen epoch and batch_size of model created via model.
        Last epoch of loss is lower than 'accept_threshold_for_loss_diff' decision be finalized.
        Epoch will test ascending format (e.g. 4, 8, 16, ..., 1024) in order to minimize time consumption
        of both parameter tuning and model creation.
        Batch size will test descending format (e.g. 1024, 512, ...4) in order to minimize time consumption
        of both parameter tuning and model creation.
        """

        counter = 0
        optimum_epoch_process_done = False
        epoch_bs_combs = list(product(sorted(self.hyper_params['batch_sizes'], reverse=True),
                                      sorted(self.hyper_params['epochs'])
                                      ))
        loss_values = []
        while not optimum_epoch_process_done:
            self.params['epochs'] = int(epoch_bs_combs[counter][1])
            self.params['batch_size'] = int(epoch_bs_combs[counter][0])
            self.build_model()
            _history = self.learning_process(save_model=False, history=True)
            if _history.history['loss'][-1] < accept_threshold_for_loss_diff:
                optimum_epoch_process_done = True
            loss_values.append({"epochs": self.params['epochs'],
                                "batch_size": self.params['batch_size'],
                                "loss": _history.history['loss'][-1]})
            counter += 1
            if counter >= parameter_tuning_trials:
                optimum_epoch_process_done = True
                loss_values_df = pd.DataFrame(loss_values).sort_values(by='loss', ascending=True)
                self.params['epochs'] = list(loss_values_df['epochs'])[0]
                self.params['batch_size'] = list(loss_values_df['batch_size'])[0]

    def remove_keras_tuner_folder(self):
        """
        removing keras tuner file. while you need to update the parameters it will affect rerun the parameter tuning.
        It won`t start unless the folder has been removed.
        """
        try:
            shutil.rmtree(join(self.directory, "untitled_project"))
        except Exception as e:
            print(" Parameter Tuning Keras Turner dummy files have already removed!!")

    def update_tuned_parameter_file(self):
        """
        tuned parameters are stored at 'test_parameters.yaml.' in given 'export_path' argument.
        """

        # existing model for 'purchase_amount' that means p. tuning was applied for both models.
        if check_for_existing_parameters(self.directory, 'purchase_amount') is not None:
            _params = read_yaml(self.directory, "test_parameters.yaml")
            _params['newcomers'] = self.params
        else:
            print("None of parameter tuning for both Model has been observed.")
            _params = {'newcomers': self.params}
        write_yaml(self.directory, "test_parameters.yaml", _params, ignoring_aliases=True)

    def parameter_tuning(self):
        if check_for_existing_parameters(self.directory, 'newcomers') is None:
            self.initialize_keras_tuner()
            self.batch_size_and_epoch_tuning()
            self.remove_keras_tuner_folder()
            self.update_tuned_parameter_file()

        else:
            self.params = check_for_existing_parameters(self.directory, 'newcomers')




